unit chargps;

{
************************************************************

chargps.pas - unit for chart.pas

(c) Heiko Purnhagen 1993-1998
e-mail: purnhage@tnt.uni-hannover.de
WWW:    http://www.fet.uni-hannover.de/~purnhage/

950803   born
960719   added GPRMC
960725   added PGRME
980420   renamed to xxGLL,... added NMEA parity check
980613   16550 (fifo) support
980614   using unit serial, timeout using 8253 cntr0
980616   nema baud rate, rx overrun handling, nmea check lf
980623   readserok if message complete / serial interrupt support

************************************************************
}


interface

uses
  serial;

const
  gpsbuflen = 128;
  gpstimeout = 500;   { 0.5 sec }

  daysec = 24.0*60.0*60.0;

  nmeatimeoutmax = 5000;   { 5 sec }
  nmeastrlen = 255;
  nmeawptlen = 8;

type
  nmeawpt = string[nmeawptlen];

var
  gpsbufidx : integer;

  gpsbuffer : array [0..gpsbuflen-1] of byte;
  gpsbufhead : integer;

  nmeastr : string[nmeastrlen];

  compolling : boolean;
  com5base : word;
  com5irq : integer;

  enable16550 : boolean;   { mainly for testing }
  fifo16550 : boolean;   { true if 16550 (fifo) detected }

  nmeatimeout : longint;

  { all gps data in rad, m, m/s }
  gpstime : real;
  gpstimeh,gpstimem,gpstimes : integer;
  gpsdatey,gpsdatem,gpsdated : integer;
  gpsx,gpsy,gpsz : real;
  gpsvele,gpsveln,gpsvelup : real;
  gpslat,gpslon,gpsalt : real;
  gpshdp,gpsvdp,gpspdp,gpstdp,gpsgdp : real;
  gpsehpe,gpsevpe,gpsete,gpsehv : real;
  gpssat : array[1..5] of record
                            track : boolean;
                            prn : integer;
                            dbhz : integer;
                          end;


function initser (com : integer; baud : longint;
                  bits, stop, parity : integer) : integer;
{ 0=error 1=ok 2=ok&fifo }
procedure restoreser;
function readserok : boolean;

function initgps (com : integer) : boolean;
{ OK->true error->false }
function readgps : integer;
function getgpsefp (idx : integer) : real;
function getgpsfp (idx : integer) : real;
function getgpsword (idx : integer) : word;
procedure getgpsdata;
function date2str (d,m,y : integer) : string;
function getgpstimestr : string;

function initnmea (com : integer; baud : longint) : boolean;
{ OK->true error->false }
function readnmea (var sent : string) : integer;
{ 0=timeout -1=error -2=check error 1=sentence w/o check 2=sent. w check }
{ check error not implemented yet }
function getnmeatype (sent : string) : integer;
{ 0=unkown 1=xxGLL 2=xxVTG 3=xxWPL 4=xxR00 5=xxRMC 6=PGRME }
function getnmeall (sent : string; var idx : integer;
                    var lat,lon : real) : boolean;
function getnmeaxxgll (sent : string; var lat,lon : real) : boolean;
{ OK->true error->false }
function getnmeaxxvtg (sent : string; var crs,spd : real) : boolean;
{ OK->true error->false }
function getnmeaxxwpl (sent : string; var lat,lon : real;
                     var name : nmeawpt) : boolean;
{ OK->true error->false }
function getnmeaxxr00 (sent : string; var idx : integer;
                     var name : nmeawpt) : boolean;
{ OK->true error->false }
{ in: idx=wpt index (1,2,...)   out: idx=num. of wtps }
function getnmeaxxrmc (sent : string; var lat,lon,crs,spd : real) : boolean;
{ OK->true error->false }
function getnmeapgrme (sent : string; var estimhorerr : real) : boolean;
{ OK->true error->false }


implementation

var
  hexstr : string[16];
  nmeamode : boolean;

{ serial interface }

function initser (com : integer; baud : longint;
                  bits, stop, parity : integer) : integer;
{ 0=error 1=ok 2=ok&fifo }
var
  fifo : boolean;
  w : word;
  l : longint;
  i : integer;
begin
  initser := 0;
  if compolling then
  begin
    sertest (com, w, fifo);
    if w = 0 then
      exit;
    serselect (com);
    if sererror then
      exit;
    serinit (baud, bits, stop, parity);
    if sererror then
      exit;
    sersetfifo (enable16550, enable16550);
    sersethandshake (false, false);
    sersettimeout (true, true, shorttimeout);
  end
  else
  begin
    if not serintinit(com, com5base, com5irq, baud, bits, stop, parity,
                      enable16550, fifo) then
      exit;
  end;
  if fifo and enable16550 then
    initser := 2   { 16550 detected }
  else
    initser := 1;
  gpsbufidx := 0;
end;

procedure restoreser;
begin
  if compolling then
    serrestore
  else
    serintrestore;
end;

function readserok : boolean;
var
  b : byte;
begin
  if gpsbufidx < 0 then
  begin
    readserok := true;
    exit;
  end;
  readserok := false;
  if compolling then
  begin
    if not serrxrdy(false) then
      exit;
    if serrxovr then
    begin
      gpsbufidx := 0;
      exit;
    end;
    b := serrx;
  end
  else
  begin
    if serintrxbuf = 0 then
      exit;
    b := serintrx;
  end;
  if serrecerr <> 0 then
  begin
    serrecerr := 0;
    gpsbufidx := 0;
    exit;
  end;
  if nmeamode then
  begin
    if gpsbufidx = 0 then
    begin
      if b = ord('$') then
      begin
        gpsbufidx := 1;
        nmeastr := '$';
      end;
    end
    else
    begin
      if gpsbufidx = nmeastrlen then
        gpsbufidx := 0
      else
        if (b = $0d) or (b = $0a) then   { cr / lf }
          gpsbufidx := -1
        else
        begin
          inc(gpsbufidx);
          nmeastr[gpsbufidx] := char(b);
          nmeastr[0] := char(gpsbufidx);
        end;
    end;
  end
  else
  begin
    if gpsbufidx < gpsbufhead then
    begin
      if b = gpsbuffer[gpsbufidx] then
        inc (gpsbufidx)
      else
        gpsbufidx := 0;
    end
    else
    begin
      gpsbuffer[gpsbufidx] := b;
      inc (gpsbufidx);
      if gpsbufidx = gpsbuflen then
        gpsbufidx := -1;
    end;
  end;
end;


{ rockwell navcore v }

function initgps (com : integer) : boolean;
{ OK->true error->false }
  var
    i : integer;
begin
  i := initser (com,9600,8,1,parityodd);
  fifo16550 := (i = 2);
  initgps := (i > 0);
  gpsbufhead := 4;
  gpsbuffer[0] := $ff;
  gpsbuffer[1] := $81;
  gpsbuffer[2] := $67;
  gpsbuffer[3] := $00;
  nmeamode := false;
end;

function readgps : integer;
{ 0 = ok  1..gpsbuflen = number of bytes left (timeout) }
  var
    t0,td : longint;
begin
  clkreset;
  t0 := clkget;
  repeat
    td := (clkget-t0) div clk1ms;
  until readserok or (td > gpstimeout);
  if gpsbufidx < 0 then
  begin
    readgps := 0;
    gpsbufidx := 0;
  end
  else
    readgps := gpsbuflen-gpsbufidx;
end;

{  (msb) s  m e (lsb) }
{ fp     1 23 8 (4 byte) }
{ efp    1 39 8 (6 byte) }
{ value = -1^s * 2^(e-129) * 1.m }

function getgpsefp (idx : integer) : real;
  var
    r : real;
begin
  r := gpsbuffer[idx+1]/256;
  r := (gpsbuffer[idx+2]+r)/256;
  r := (gpsbuffer[idx+3]+r)/256;
  r := (gpsbuffer[idx+4]+r)/256;
  r := 1+((gpsbuffer[idx+5] and $7f)+r)/128;
  if gpsbuffer[idx+5] and $80 = $80 then
    r := -r;
  getgpsefp := r*exp((gpsbuffer[idx]-129)*ln(2));
end;

function getgpsfp (idx : integer) : real;
  var
    r : real;
begin
  r := gpsbuffer[idx+1]/256;
  r := (gpsbuffer[idx+2]+r)/256;
  r := 1+((gpsbuffer[idx+3] and $7f)+r)/128;
  if gpsbuffer[idx+3] and $80 = $80 then
    r := -r;
  getgpsfp := r*exp((gpsbuffer[idx]-129)*ln(2));
end;

function getgpsword (idx : integer) : word;
{ ls byte first }
begin
  getgpsword := gpsbuffer[idx]+256*gpsbuffer[idx+1];
end;

procedure getgpsdata;
  var
    time : longint;
    i : integer;
begin
  gpstime := getgpsefp(20);
  time := round(gpstime);
  gpstimes := time mod 60;
  time := time div 60;
  gpstimem := time mod 60;
  gpstimeh := time div 60;
  gpsdated := getgpsword(26);
  gpsdatem := getgpsword(28);
  gpsdatey := getgpsword(30);
  gpsx := getgpsefp(36);
  gpsy := getgpsefp(42);
  gpsz := getgpsefp(48);
  gpsvele := getgpsfp(54);
  gpsveln := getgpsfp(58);
  gpsvelup := getgpsfp(62);
  gpslat := getgpsefp(66);
  gpslon := getgpsefp(72);
  gpsalt := getgpsfp(78);
  gpshdp := getgpsword(82) / 100;
  gpsvdp := getgpsword(84) / 100;
  gpspdp := getgpsword(86) / 100;
  gpstdp := getgpsword(88) / 100;
  gpsgdp := getgpsword(90) / 100;
  gpsehpe := getgpsword(112);
  gpsevpe := getgpsword(114);
  gpsete := getgpsword(116) / 300e6;
  gpsehv := getgpsword(118) / 100;
  for i := 1 to 5 do
  begin
    gpssat[i].track := ((gpsbuffer[93+(i-1)*4] and $8) = $8);
    gpssat[i].prn := gpsbuffer[92+(i-1)*4];
    gpssat[i].dbhz := gpsbuffer[95+(i-1)*4];
  end;
end;

function date2str (d,m,y : integer) : string;
  var
    s,ss : string;
begin
  str (d:2,ss);
  if ss[1]=' ' then
    ss[1] := '0';
  s := ss+'-';
  ss := '';
  case m of
     1 : ss := 'jan';
     2 : ss := 'feb';
     3 : ss := 'mar';
     4 : ss := 'apr';
     5 : ss := 'may';
     6 : ss := 'jun';
     7 : ss := 'jul';
     8 : ss := 'aug';
     9 : ss := 'sep';
    10 : ss := 'oct';
    11 : ss := 'nov';
    12 : ss := 'dec';
  end;
  s := s+ss+'-';
  str (y:4,ss);
  date2str := s+ss;
end;

function getgpstimestr : string;
  var
    s,ss : string;
begin
  s := date2str(gpsdated,gpsdatem,gpsdatey);
  s := s+' ';
  str (gpstimeh:2,ss);
  if ss[1]=' ' then
    ss[1] := '0';
  s := s+ss+':';
  str (gpstimem:2,ss);
  if ss[1]=' ' then
    ss[1] := '0';
  s := s+ss+':';
  str (gpstimes:2,ss);
  if ss[1]=' ' then
    ss[1] := '0';
  s := s+ss+' ';
  getgpstimestr := s;
end;


{ nmea0183   03-aug-95 }

function initnmea (com : integer; baud : longint) : boolean;
{ OK->true error->false }
  var
    i : integer;
begin
  i := initser (com,baud,8,1,parityno);
  if baud >= 2400 then
    nmeatimeout := 500
  else
    nmeatimeout := 1000;
  fifo16550 := (i = 2);
  initnmea := (i > 0);
  nmeamode := true;
end;

function readnmea (var sent : string) : integer;
{ 0=timeout -1=error -2=check error 1=sentence w/o check 2=sent. w check }
  var
    t0,td : longint;
    i,j : integer;
    b : byte;
begin
  clkreset;
  t0 := clkget;
  repeat
    td := (clkget-t0) div clk1ms;
  until readserok or (td > nmeatimeout);
  if gpsbufidx >= 0 then
  begin
    readnmea := 0;
    exit;
  end;
  gpsbufidx := 0;
  sent := nmeastr;
  i := ord(sent[0]);
  if i < 6 then
  begin
    readnmea := -1;
    exit;
  end;
  if sent[i-2] = '*' then
  begin
    readnmea := 2;   { sentence with checksum }
    sent[0] := chr(i-3);   { set to length without checksum }
    b := 0;
    for j := 2 to i-3 do
      b := b xor ord(sent[j]);
    if (sent[i-1] <> hexstr[((b shr 4) and $0f)+1]) or
       (sent[i] <> hexstr[(b and $0f)+1]) then
      readnmea := -2;   { checksum error }
  end
  else
    readnmea := 1;   { sentence without checksum }
end;

function getnmeatype (sent : string) : integer;
{ 0=unkown 1=xxGLL 2=xxVTG 3=xxWPL 4=xxR00 5=xxRMC 6=PGRME }
  var
    t : string[5];
begin
  getnmeatype := 0;
  t := copy(sent,4,3);
  if t='GLL' then
    getnmeatype := 1;
  if t='VTG' then
    getnmeatype := 2;
  if t='WPL' then
    getnmeatype := 3;
  if t='R00' then
    getnmeatype := 4;
  if t='RMC' then
    getnmeatype := 5;
  t := copy(sent,2,5);
  if t='PGRME' then
    getnmeatype := 6;
end;

function getnmeall (sent : string; var idx : integer;
                    var lat,lon : real) : boolean;
{ OK->true error->false }
  var
    hh,mm : real;
    c,i : integer;
begin
  getnmeall := false;
  if length(sent) < idx+1 then
    exit;
  val(copy(sent,idx,2),hh,c);
  if c <> 0 then
    exit;
  idx := idx+2;
  i := idx;
  while (i <= length(sent)) and (sent[i] <> ',') do
    inc (i);
  if i > length(sent) then
    exit;
  val(copy(sent,idx,i-idx),mm,c);
  if c <> 0 then
    exit;
  idx := i+1;
  if length(sent) < idx then
    exit;
  if (sent[idx] <> 'N') and (sent[idx] <> 'S') then
    exit;
  lat := (hh+mm/60)/180*pi;
  if sent[idx] = 'S' then
    lat := -lat;
  inc (idx);
  if (length(sent) < idx) or (sent[idx] <> ',') then
    exit;
  inc (idx);
  if length(sent) < idx+2 then
    exit;
  val(copy(sent,idx,3),hh,c);
  if c <> 0 then
    exit;
  idx := idx+3;
  i := idx;
  while (i <= length(sent)) and (sent[i] <> ',') do
    inc (i);
  if i > length(sent) then
    exit;
  val(copy(sent,idx,i-idx),mm,c);
  if c <> 0 then
    exit;
  idx := i+1;
  if length(sent) < idx then
    exit;
  if (sent[idx] <> 'E') and (sent[idx] <> 'W') then
    exit;
  lon := (hh+mm/60)/180*pi;
  if sent[idx] = 'W' then
    lon := -lon;
  inc (idx);
  getnmeall := true;
end;

function getnmeaxxgll (sent : string; var lat,lon : real) : boolean;
{ OK->true error->false }
  var
    i : integer;
begin
  getnmeaxxgll := false;
  i := 7;
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  getnmeaxxgll := getnmeall(sent,i,lat,lon);
end;

function getnmeaxxvtg (sent : string; var crs,spd : real) : boolean;
{ OK->true error->false }
  var
    i,j,c : integer;
begin
  getnmeaxxvtg := false;
  i := 7;
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  j := i;
  while (j <= length(sent)) and (sent[j] <> ',') do
    inc (j);
  if j > length(sent) then
    exit;
  if j = i then
    crs := 0
  else
  begin
    val(copy(sent,i,j-i),crs,c);
    if c <> 0 then
      exit;
    crs := crs/180*pi;
  end;
  i := j+1;
  if length(sent) < i then
    exit;
  if sent[i] <> 'T' then
    exit;
  inc (i);
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  { skip mag crs }
  while (i <= length(sent)) and (sent[i] <> ',') do
    inc (i);
  if i > length(sent) then
    exit;
  inc (i);
  { skip M }
  while (i <= length(sent)) and (sent[i] <> ',') do
    inc (i);
  if i > length(sent) then
    exit;
  inc (i);
  j := i;
  while (j <= length(sent)) and (sent[j] <> ',') do
    inc (j);
  if j > length(sent) then
    exit;
  val(copy(sent,i,j-i),spd,c);
  if c <> 0 then
    exit;
  i := j+1;
  if length(sent) < i then
    exit;
  if sent[i] <> 'N' then
    exit;
  getnmeaxxvtg := true;
end;

function getnmeaxxwpl (sent : string; var lat,lon : real;
                     var name : nmeawpt) : boolean;
{ OK->true error->false }
  var
    i,j : integer;
begin
  getnmeaxxwpl := false;
  name := '';
  i := 7;
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  if not getnmeall(sent,i,lat,lon) then
    exit;
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  j := i;
  while (j <= length(sent)) and (sent[j] <> ',') do
    inc (j);
  j := j-i;
  if j > nmeawptlen then
    j := nmeawptlen;
  name := copy(sent,i,j);
  getnmeaxxwpl := true;
end;

function getnmeaxxr00 (sent : string; var idx : integer;
                     var name : nmeawpt) : boolean;
{ OK->true error->false }
{ in: idx=wpt index (1,2,...)   out: idx=num. of wtps }
  var
    i,j,k,n : integer;
begin
  getnmeaxxr00 := false;
  name := '';
  i := 7;
  n := 0;
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  repeat
    inc (n);
    j := i;
    while (j <= length(sent)) and (sent[j] <> ',') do
      inc (j);
    if n = idx then
    begin
      k := j-i;
      if k > nmeawptlen then
        k := nmeawptlen;
      name := copy(sent,i,k);
      getnmeaxxr00 := true;
    end;
    i := j;
    if (i > length(sent)) then
    begin
      idx := n;
      exit;
    end;
    inc (i);
    if (i > length(sent)) or (sent[i] = ',') then
    begin
      idx := n;
      exit;
    end;
  until false;
end;

function getnmeaxxrmc (sent : string; var lat,lon,crs,spd : real) : boolean;
{ OK->true error->false }
  var
    i,j,c : integer;
begin
  getnmeaxxrmc := false;
  i := 7;
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  { skip hhmmss }
  while (i <= length(sent)) and (sent[i] <> ',') do
    inc (i);
  if i > length(sent) then
    exit;
  inc (i);
  { skip A }
  while (i <= length(sent)) and (sent[i] <> ',') do
    inc (i);
  if i > length(sent) then
    exit;
  inc (i);
  if not getnmeall(sent,i,lat,lon) then
    exit;
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  j := i;
  while (j <= length(sent)) and (sent[j] <> ',') do
    inc (j);
  if j > length(sent) then
    exit;
  val(copy(sent,i,j-i),spd,c);
  if c <> 0 then
    exit;
  i := j+1;
  if length(sent) < i then
    exit;
  j := i;
  while (j <= length(sent)) and (sent[j] <> ',') do
    inc (j);
  if j > length(sent) then
    exit;
  if j = i then
    crs := 0
  else
  begin
    val(copy(sent,i,j-i),crs,c);
    if c <> 0 then
      exit;
    crs := crs/180*pi;
  end;
  { ignore ddmmyy & deviation }
  getnmeaxxrmc := true;
end;

function getnmeapgrme (sent : string; var estimhorerr : real) : boolean;
{ OK->true error->false }
  var
    i,j,c : integer;
begin
  getnmeapgrme := false;
  i := 7;
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  j := i;
  while (j <= length(sent)) and (sent[j] <> ',') do
    inc (j);
  if j > length(sent) then
    exit;
  if j = i then
    estimhorerr := 0
  else
  begin
    val(copy(sent,i,j-i),estimhorerr,c);
    if c <> 0 then
      exit;
  end;
  i := j+1;
  if length(sent) < i then
    exit;
  if sent[i] <> 'M' then
    exit;
  getnmeapgrme := true;
end;


begin
  hexstr := '0123456789ABCDEF';
  nmeamode := false;
  nmeatimeout := nmeatimeoutmax;
  compolling := false;
  com5base := $3f8;
  com5irq := 4;
  enable16550 := true;
end.

{ end unit }
