unit chartran;

{
************************************************************

chartran.pas - unit for chart.pas

(c) Heiko Purnhagen 1993-1997
e-mail: purnhage@tnt.uni-hannover.de
WWW:    http://www.fet.uni-hannover.de/purnhage/

940103   born
960723   ...
970729   ...
970811   ...

************************************************************
}

interface

const
  r = 3437.7468; { earth radius: 360*60/(2*pi) nm }
  nm = 1851.85;  { 1 nm = 40000e3/(360*60) m }
  s0 = 1;        { scale at lat=0 }

type
  llt = record
          lat,lon : real;
        end;
  xyt = record
          x,y : real;
        end;
  uvt = record
          u,v : real;
        end;
  llart = array[1..3] of llt;
  xyart = array[1..3] of xyt;
  uvart = array[1..3] of uvt;

var
  a11,a12,a21,a22 : real;
  b1,b2 : real;
  sa11,sa12,sa21,sa22 : real;
  sb1,sb2 : real;
  tmpa11,tmpa12,tmpa21,tmpa22 : real;
  tmpb1,tmpb2 : real;
  tllformat : integer;
  tlllength : integer;

function tan (x : real) : real;
function calcang (x,y : real) : real;
{ angle in ]-pi,pi] }
procedure calcdistbear (ll1,ll2 : llt; var dist,brg : real);
{ rhumb line }
{ distance not precise if big !!! }
procedure ll2xy (ll : llt; var xy : xyt);
procedure xy2ll (xy : xyt; var ll : llt);
procedure xy2uv (xy : xyt; var uv : uvt);
procedure uv2xy (uv : uvt; var xy : xyt);
function ang2str (ang : real; h3 : boolean; len : integer) : string;
function ll2str (ll : llt) : string;
function str2ang (s : string; var ang : real) : boolean;
function str2ll (s : string; var ll : llt) : boolean;
function findab (xy : xyart; uv : uvart) : boolean;
procedure saveab;
procedure restoreab;
procedure saveabtmp;
procedure restoreabtmp;
procedure initll2str (llformat,lllength : integer);


implementation

{ procedures for coordinate transforms }

function tan (x : real) : real;
begin
  tan := sin(x)/cos(x);
end;

function calcang (x,y : real) : real;
{ angle in ]-pi,pi] }
begin
  if x = 0 then
  begin
    if y = 0 then
      calcang := 0
    else
      if y > 0 then
        calcang := pi/2
      else
        calcang := -pi/2;
  end
  else
    if x > 0 then
      calcang := arctan(y/x)
    else
      if y >= 0 then
        calcang := arctan(y/x)+pi
      else
        calcang := arctan(y/x)-pi;
end;

procedure calcdistbear (ll1,ll2 : llt; var dist,brg : real);
{ rhumb line }
{ distance not precise if big !!! }
  var
    xy1,xy2 : xyt;
    lat : real;
begin
  ll2xy (ll1,xy1);
  ll2xy (ll2,xy2);
  lat := (ll1.lat+ll2.lat)/2;
  dist := sqrt(sqr(xy2.x-xy1.x)+sqr(xy2.y-xy1.y))*cos(lat);
  if dist = 0 then
    brg := 0
  else
  begin
    brg := calcang (xy2.y-xy1.y,xy2.x-xy1.x);
    if brg < 0 then
      brg := brg+2*pi;
  end;
end;

procedure ll2xy (ll : llt; var xy : xyt);
begin
  xy.x := r*s0*ll.lon;
  xy.y := r*s0*ln(tan(ll.lat/2+pi/4));
end;

procedure xy2ll (xy : xyt; var ll : llt);
begin
  ll.lat := 2*arctan(exp(xy.y/(r*s0)))-pi/2;
  ll.lon := xy.x/(r*s0);
end;

procedure xy2uv (xy : xyt; var uv : uvt);
  var
    det,x,y : real;
begin
  det := a11*a22-a12*a21;
  x := xy.x-b1;
  y := xy.y-b2;
  uv.u := (a22*x-a12*y)/det;
  uv.v := (-a21*x+a11*y)/det;
end;

procedure uv2xy (uv : uvt; var xy : xyt);
begin
  xy.x := a11*uv.u+a12*uv.v+b1;
  xy.y := a21*uv.u+a22*uv.v+b2;
end;

function ang2str (ang : real; h3 : boolean; len : integer) : string;
  var
    h,m,s : real;
    sh,sm,ss : string;
    l,h1,h2,m1,m2,s1,s2 : integer;
begin
  h := ang/pi*180;
  sm := '';
  ss := '';
  if h3 then
    l := 3
  else
    l := 2;
  h1 := l;
  h2 := 0;
  if tllformat = 1 then
  begin
    if len >= l+3 then
    begin
      h1 := len-1;
      h2 := len-l-2;
    end;
  end
  else
  begin
    m := frac(h)*60.0;
    h := trunc(h);
  end;
  if tllformat <> 1 then
  begin
    len := len-l-1;
    m1 := 2;
    m2 := 0;
    if tllformat <> 3 then
    begin
      if len >= 5 then
      begin
        m1 := len-1;
        m2 := len-4;
      end;
    end
    else
    begin
      s := frac(m)*60.0;
      m := trunc(m);
    end;
    if tllformat = 3 then
    begin
      len := len-3;
      s1 := 2;
      s2 := 0;
      if len >= 5 then
      begin
        s1 := len-1;
        s2 := len-4;
      end;
      str (s:s1:s2,ss);
      if ss[1] = '6' then
      begin
        str (0.0:s1:s2,ss);
        m := m+1;
      end;
      if ss[1] = ' ' then
        ss[1] := '0';
      ss := ss+'"';
    end;
    str (m:m1:m2,sm);
    if sm[1] = '6' then
    begin
      str (0.0:m1:m2,sm);
      h := h+1;
    end;
    if sm[1] = ' ' then
      sm[1] := '0';
    sm := sm+'''';
  end;
  str (h:h1:h2,sh);
  if l = 3 then
    if sh[2] = ' ' then
      sh[2] := '0';
  if sh[1] = ' ' then
    sh[1] := '0';
  sh := sh+'';
  ang2str := sh+sm+ss;
end;

function ll2str (ll : llt) : string;
  var
    s,ss : string;
    lt : real;
    len : integer;
begin
  len := tlllength div 2;
  if ll.lat >= 0 then
  begin
    lt := ll.lat;
    ss := 'N';
  end
  else
  begin
    lt := -ll.lat;
    ss := 'S';
  end;
  s := ang2str(lt,false,len-2)+ss;
  if ll.lon >= 0 then
  begin
    lt := ll.lon;
    ss := 'E';
  end
  else
  begin
    lt := -ll.lon;
    ss := 'W';
  end;
  ll2str := s+' '+ang2str(lt,true,len-1)+ss;
end;

function str2ang (s : string; var ang : real) : boolean;
{ OK -> true, ERROR -> false}
  const
    tab = #9;
  var
    ts : string;
    hh,mm,ss : real;
    p,c : integer;
begin
  str2ang := false;
  hh := 0;
  mm := 0;
  ss := 0;
  while (s <> '') and (s[length(s)] in [' ',tab]) do
    delete (s,length(s),1);
  while (s <> '') and (s[1] in [' ',tab]) do
    delete (s,1,1);
  if s = '' then
    exit;
  p := pos('',s);
  if p = 0 then
    p := pos(':',s);
  if p = 0 then
    p := pos(' ',s);
  if p = 0 then
    ts := s
  else
    ts := copy(s,1,p-1);
  val (ts,hh,c);
  if c <> 0 then
    exit;
  if hh < 0 then
    exit;
  if p > 0 then
  begin
    delete (s,1,p);
    while (s <> '') and (s[1] in [' ',tab]) do
      delete (s,1,1);
    if s <> '' then
    begin
      p := pos('''',s);
      if p = 0 then
        p := pos(':',s);
      if p = 0 then
        p := pos(' ',s);
      if p = 0 then
        ts := s
      else
        ts := copy(s,1,p-1);
      val (ts,mm,c);
      if c <> 0 then
        exit;
      if (frac(hh) <> 0) or (mm < 0) or (mm > 60) then
        exit;
      if p > 0 then
      begin
        delete (s,1,p);
        while (s <> '') and (s[1] in [' ',tab]) do
          delete (s,1,1);
        if s <> '' then
        begin
          p := pos('"',s);
          if p = 0 then
            p := pos(' ',s);
          if p = 0 then
            ts := s
          else
            ts := copy(s,1,p-1);
          val (ts,ss,c);
          if c <> 0 then
            exit;
          if (frac(mm) <> 0) or (ss < 0) or (ss > 60) then
            exit;
          if p > 0 then
          begin
            delete (s,1,p);
            while (s <> '') and (s[1] in [' ',tab]) do
              delete (s,1,1);
            if s <> '' then
              exit;
          end;
        end;
      end;
    end;
  end;
  ang := (hh+mm/60+ss/3600)/180*pi;
  str2ang := true;
end;

function str2ll (s : string; var ll : llt) : boolean;
{ OK -> true, ERROR -> false}
  var
    ss : string;
    sign : integer;
    p1,p2 : integer;
begin
  str2ll := false;
  p1 := pos('N',s);
  if p1 = 0 then
    p1 := pos('n',s);
  sign := 1;
  if p1 = 0 then
  begin
    p1 := pos('S',s);
    if p1 = 0 then
      p1 := pos('s',s);
    sign := -1;
  end;
  if p1 = 0 then
    exit;
  if not str2ang(copy(s,1,p1-1),ll.lat) then
      exit;
  ll.lat := sign*ll.lat;
  if (ll.lat > pi/2) or (ll.lat < -pi/2) then
    exit;
  p2 := pos('E',s);
  if p2 = 0 then
    p2 := pos('e',s);
  sign := 1;
  if p2 = 0 then
  begin
    p2 := pos('W',s);
    if p2 = 0 then
      p2 := pos('w',s);
    sign := -1;
  end;
  if p2 = 0 then
    exit;
  if not str2ang(copy(s,p1+1,p2-p1-1),ll.lon) then
      exit;
  ll.lon := sign*ll.lon;
  if (ll.lon > 2*pi) or (ll.lon < -2*pi) then   { avoids problems at 180 }
    exit;
  str2ll := true;
end;

function findab (xy : xyart; uv : uvart) : boolean;
{ OK -> true, ERROR -> false}
  var
    det,a11t,a12t,a21t,a22t : real;
begin
  findab := false;
  det :=  uv[1].u*uv[2].v + uv[2].u*uv[3].v + uv[3].u*uv[1].v
        - uv[1].u*uv[3].v - uv[2].u*uv[1].v - uv[3].u*uv[2].v;
  if det = 0 then
    exit;
  a11t := (xy[1].x*uv[2].v + xy[2].x*uv[3].v + xy[3].x*uv[1].v
         - xy[1].x*uv[3].v - xy[2].x*uv[1].v - xy[3].x*uv[2].v) / det;
  a12t := (uv[1].u*xy[2].x + uv[2].u*xy[3].x + uv[3].u*xy[1].x
         - uv[1].u*xy[3].x - uv[2].u*xy[1].x - uv[3].u*xy[2].x) / det;
  a21t := (xy[1].y*uv[2].v + xy[2].y*uv[3].v + xy[3].y*uv[1].v
         - xy[1].y*uv[3].v - xy[2].y*uv[1].v - xy[3].y*uv[2].v) / det;
  a22t := (uv[1].u*xy[2].y + uv[2].u*xy[3].y + uv[3].u*xy[1].y
         - uv[1].u*xy[3].y - uv[2].u*xy[1].y - uv[3].u*xy[2].y) / det;
  if a11t*a22t-a12t*a21t = 0 then   { det in xy2uv }
    exit;
  a11 := a11t;
  a12 := a12t;
  a21 := a21t;
  a22 := a22t;
  b1  := (uv[1].u*uv[2].v*xy[3].x + uv[2].u*uv[3].v*xy[1].x
        + uv[3].u*uv[1].v*xy[2].x
        - uv[1].u*uv[3].v*xy[2].x - uv[2].u*uv[1].v*xy[3].x
        - uv[3].u*uv[2].v*xy[1].x) / det;
  b2  := (uv[1].u*uv[2].v*xy[3].y + uv[2].u*uv[3].v*xy[1].y
        + uv[3].u*uv[1].v*xy[2].y
        - uv[1].u*uv[3].v*xy[2].y - uv[2].u*uv[1].v*xy[3].y
        - uv[3].u*uv[2].v*xy[1].y) / det;
  findab := true;
end;

procedure saveab;
begin
  sa11 := a11;
  sa12 := a12;
  sa21 := a21;
  sa22 := a22;
  sb1 := b1;
  sb2 := b2;
end;

procedure restoreab;
begin
  a11 := sa11;
  a12 := sa12;
  a21 := sa21;
  a22 := sa22;
  b1 := sb1;
  b2 := sb2;
end;

procedure saveabtmp;
begin
  tmpa11 := a11;
  tmpa12 := a12;
  tmpa21 := a21;
  tmpa22 := a22;
  tmpb1 := b1;
  tmpb2 := b2;
end;

procedure restoreabtmp;
begin
  a11 := tmpa11;
  a12 := tmpa12;
  a21 := tmpa21;
  a22 := tmpa22;
  b1 := tmpb1;
  b2 := tmpb2;
end;

procedure initll2str (llformat,lllength : integer);
{ llformat: 1 = H.h }
{           2 = H M.m }
{           3 = H M S.s }
{ lllength: length of ll string (#chars) }
begin
  tllformat := llformat;
  tlllength := lllength;
end;

begin
end.

{ end unit }
